# SPDX-FileCopyrightText: 2020 Jim Bennett for Adafruit Industries
# SPDX-FileCopyrightText: 2020 Elena Horton for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`constants`
================================================================================

This file is for maintaining Microsoft Azure IoT constants that could be changed or added to over time for different scenarios

* Author(s): Jim Bennett, Elena Horton
"""

# The version of the IoT Central MQTT API this code is built against
IOTC_API_VERSION = "2019-10-01"

# The version of the Azure Device Provisioning Service this code is built against
DPS_API_VERSION = "2019-03-31"

# The Azure Device Provisioning service endpoint that this library uses to provision IoT Central devices
DPS_END_POINT = "global.azure-devices-provisioning.net"
